import React, { useEffect, useMemo, useReducer, useRef, createContext, useContext } from "react";
import { tryCreateDevToolsLogger } from "./devtools";
const keysOf = (obj) => Object.keys(obj);
export function useMethods(initialState, methods) {
    // To avoid warnings "Can't perform a React state update on an unmounted component"
    const isMounted = useRef(true);
    useEffect(() => () => {
        isMounted.current = false;
    }, []);
    const reducer = useMemo(() => {
        const logger = tryCreateDevToolsLogger();
        return (state, action) => {
            const nextState = methods[action.type](state)(...action.payload);
            if (logger) {
                logger(action, nextState);
            }
            return nextState;
        };
    }, []);
    const [state, dispatch] = useReducer(reducer, initialState);
    const methodsOut = useMemo(() => {
        const methodsOut = {};
        keysOf(methods).forEach(type => {
            methodsOut[type] = (...payload) => {
                // Do not dispatch actions on unmounted Provider
                if (!isMounted.current) {
                    return;
                }
                dispatch({
                    type,
                    payload
                });
            };
        });
        return methodsOut;
    }, []);
    return [state, methodsOut];
}
export class NoProviderError extends Error {
}
const createStubMethods = (methodsIn) => keysOf(methodsIn).reduce((methods, type) => {
    methods[type] = () => {
        throw new NoProviderError("Store method called without top-level Provider");
    };
    return methods;
}, {});
export const createUseStore = (defaultState, methodsIn) => {
    const methodsStubsWithoutProvider = createStubMethods(methodsIn);
    const context = createContext([
        defaultState,
        methodsStubsWithoutProvider
    ]);
    const StoreProvider = ({ initialState, children }) => {
        const [state, methods] = useMethods(initialState, methodsIn);
        return (React.createElement(context.Provider, { value: [state, methods] }, children));
    };
    const useStore = () => {
        const [state, methods] = useContext(context);
        return { state, methods };
    };
    useStore.Provider = StoreProvider;
    return useStore;
};
